#include <R.h>
#include <Rmath.h>
#include <Rinternals.h>
#include <R_ext/Lapack.h>
#include <R_ext/BLAS.h>

#ifdef _OPENMP
#include <omp.h>
#endif

//Description: returns the lower triangular plus diagonal portion of the model covariance matrix Sigma_{y|theta} = sigma^2 R(phi) + tau^2 I
//C - nxn matrix that holds the lower triangle plus diagonal of the correlation matrix on exit (must be pre-allocated)
//coords - nx2 matrix of spatial coordinates
//n - number of rows in coords, and rows/columns of R
//tauSq - measurement variance (nugget)
//sigmaSq - spatial variance (partial sill)
//phi - spatial decay parameter

void sigma(double *C, double* coords, int n, double tauSq, double sigmaSq, double phi){
  
  int i,j;
  
  //create the correlation matrix
  for(i = 0; i < n; i++){
    for(j = 0; j <= i; j++){
      C[j*n+i] = sigmaSq*exp(-phi*sqrt(pow(coords[i]-coords[j], 2)+pow(coords[n+i]-coords[n+j], 2)));
      if(i == j){
	C[i*n+i] += tauSq;
      }
    }
  }
}


//Description: generates random numbers from a n-dimensional multivariate normal N(mu, Sigma)
//des - pre-allocated vector to receive the n random numbers
//mu - mean vector
//L - lower triangle of chol(Sigma), i.e., Sigma = LL', output from dpotrf
//n - length of mu and rows and columns of L 
void mvrnorm(double *des, double *mu, double *L, int n){
  
  int i;
  int inc = 1;
  double one = 1.0;
  double zero = 0.0;
  char const *lower = "L";
  char const *nDiag = "N";
  char const *nTran = "N";

  //make some std norm draws
  for(i = 0; i < n; i++){
    des[i] = rnorm(0.0, 1.0);
  }
  
  //mult this vector by L
  F77_NAME(dtrmv)(lower, nTran, nDiag, &n, L, &n, des, &inc);
  
  //add the mean to the result
  F77_NAME(daxpy)(&n, &one, mu, &inc, des, &inc);
  
}


//Description: prints the nxp matrix to the R console (helps with debugging)
void show(double *a, int n, int p){

  for(int i = 0; i < n; i++){
    for(int j = 0; j < p; j++){
      Rprintf("%.5f\t", a[j*n+i]);
    }
    Rprintf("\n");
  }
}

//Description: zeros out vector a of length n
void zeros(double *a, int n){
  for(int i = 0; i < n; i++){
    a[i] = 0;
  }
}

//Description: zeros out the upper triangle (not including diagonal) of a square nxn matix
void zerosUpTri(double *a, int n){
  int i, j;

  for(i = 0; i < n; i++){
    for(j = 0; j < i; j++){
      a[i*n+j] = 0;
    }
  }  
  
  
}


extern "C" {
  
  SEXP collaspedGPLM(SEXP y_r, SEXP X_r, SEXP n_r, SEXP p_r, SEXP coords_r, 
		     SEXP tauSq_r, SEXP sigmaSq_r, SEXP phi_r,
		     SEXP nSamples_r, SEXP nThreads_r){

    //////////////////////////////////
    //get input arguments
    //////////////////////////////////
    //data
    double *y = REAL(y_r);
    double *X = REAL(X_r);
    int p = INTEGER(p_r)[0];
    int n = INTEGER(n_r)[0];
    double *coords = REAL(coords_r);

    //fixed parameter values
    double sigmaSq = REAL(sigmaSq_r)[0];
    double tauSq = REAL(tauSq_r)[0];
    double phi = REAL(phi_r)[0];

    //other stuff
    int nSamples = INTEGER(nSamples_r)[0];
    int nThreads = INTEGER(nThreads_r)[0];

    ////////////////////////////////////////////
    //define some stuff we might use
    ///////////////////////////////////////////
    int i, j, info, nProtect=0;
    const int inc = 1;
    const double one = 1.0;
    const double negOne = -1.0;
    const double zero = 0.0;
    char const *lower = "L";
    char const *nUnit = "N";
    char const *nTran = "N";
    char const *yTran = "T";
    char const *rSide = "R";
    char const *lSide = "L";
    char const *nDiag = "N";
    int np = n*p;
    int pp = p*p;
    int nn = n*n;

   GetRNGstate();

    
    ////////////////////////////////////////////
    //sample beta's posterior 
    ////////////////////////////////////////////
    //assuming B's prior is flat mu_beta = 0 and Sigma_beta = 0
    
    //the samples of beta will be passed back to the calling program as a R matrix allocated here
    SEXP betaSamples_r ;
    PROTECT(betaSamples_r = allocMatrix(REALSXP, p, nSamples)); nProtect++;
    
    //allocate space for components we need
    double *L = (double *) R_alloc(n*n, sizeof(double));
    double *vU = (double *) R_alloc(n+np, sizeof(double));
    double *b = (double *) R_alloc(p, sizeof(double));
    double *L_B = (double *) R_alloc(pp, sizeof(double));
    int p1 = p+1;// number of columns of vU which we'll need for the call to dtrsm below
    
    //get L  = chol(Sigma_{y|beta,theta}), I'll just start with L = Sigma_{y|beta,theta} then the desired L will be left after calling dpotrf
    sigma(L, coords, n, tauSq, sigmaSq, phi); //fill in the lower triangle of the covariance matrix
    
    F77_NAME(dpotrf)(lower, &n, L, &n, &info); if(info != 0){error("c++ error: dpotrf failed\n");}
    
    //solve both Lv=y and LU=X with one call to dtrsm, the result will be stored in vU
    F77_NAME(dcopy)(&n, y, &inc, vU, &inc); //copy y into first column of vU, i.e., v
    
    F77_NAME(dcopy)(&np, X, &inc, &vU[n], &inc); //copy X into the subsequent columns of vU, i.e., U
    
    F77_NAME(dtrsm)(lSide, lower, nTran, nUnit, &n, &p1, &one, L, &n, vU, &n);//solve L[v:U] = [y:X] 

    //get b
    F77_NAME(dgemv)(yTran, &n, &p, &one, &vU[n], &n, vU, &inc, &zero, b, &inc); //U'v

    //solve in succession trisolve(L'_B, trisolve(L_B, b)), where L_B = chol(U'U) which we get first
    F77_NAME(dgemm)(yTran, nTran, &p, &p, &n, &one, &vU[n], &n, &vU[n], &n, &zero, L_B, &p); //U'U
    F77_NAME(dpotrf)(lower, &p, L_B, &p, &info); if(info != 0){error("c++ error: dpotrf failed\n");}

    //solve trisolve(L_B, b)  L_Bx = b, where b = x on exit 
    F77_NAME(dtrsv)(lower, nTran, nUnit, &p, L_B, &p, b, &inc);
    
    //now solve trisolve(L_B', x)  L_Bv = x, where b = v on exit 
    F77_NAME(dtrsv)(lower, yTran, nUnit, &p, L_B, &p, b, &inc);

    for(i = 0; i < nSamples; i++){

      //make z then do L_B z and save the result in the sample matrix column
      for(j = 0; j < p; j++){
	REAL(betaSamples_r)[i*p+j] = rnorm(0.0, 1.0);
      }

      //result left in place
      F77_NAME(dtrsv)(lower, yTran, nUnit, &p, L_B, &p, &REAL(betaSamples_r)[i*p], &inc);
      
      //add trisolve(L'_B, trisolve(L_B, b)) held in b to result from above
      F77_NAME(daxpy)(&p, &one, b, &inc, &REAL(betaSamples_r)[i*p], &inc);
      
    }

    ////////////////////////////////////////////
    //sample w's posterior 
    ////////////////////////////////////////////
    //the samples of w will be passed back to the calling program as a R matrix which we allocate here
    SEXP wSamples_r;
    PROTECT(wSamples_r = allocMatrix(REALSXP, n, nSamples)); nProtect++;

    //allocate space for components we need
    double *W = (double *) R_alloc(nn, sizeof(double));
    double *L_A = (double *) R_alloc(nn, sizeof(double));
  
    //following the idenity (K(theta)^{-1} + G(theta)^{-1})^{-1} = G(theta) - G(theta)(K(theta) + G(theta))^{-1}G(theta),
    //our is G = tau^sq I since we're assuming a flat prior on beta

    //get L = chol(K(theta) + G), again just start with L = (K(theta) + G) then the desired L will be left after calling dpotrf
    sigma(L, coords, n, tauSq, sigmaSq, phi); //fill in the lower triangle of the covariance matrix
    
    F77_NAME(dpotrf)(lower, &n, L, &n, &info); if(info != 0){error("c++ error: dpotrf failed\n");}
    
    //solve L W = G to get W. So set copy W = G then the result of dtrsm of be left in W
    zeros(W, nn);//zero out W

    for(i = 0; i < n; i++){
      W[i*n+i] = tauSq;
    }

    F77_NAME(dtrsm)(lSide, lower, nTran, nUnit, &n, &n, &one, L, &n, W, &n);

    //get L_A, again initially set L_A to G - W'W then L_A will be left after the call to dpotrf
    F77_NAME(dgemm)(yTran, nTran, &n, &n, &n, &negOne, W, &n, W, &n, &zero, L_A, &n);//puts -W'W in L_A
    
    for(i = 0; i < n; i++){
      L_A[i*n+i] += tauSq;
    }
    
    F77_NAME(dpotrf)(lower, &n, L_A, &n, &info); if(info != 0){error("c++ error: dpotrf failed\n");}

    for(i = 0; i < nSamples; i++){

      //make L_A L_A' b where b is (y - X beta)/tau^2
      //I don't want to save over y so copy it into v (where v was allocated above for beta sampling)
      F77_NAME(dgemv)(nTran, &n, &p, &negOne, X, &n, &REAL(betaSamples_r)[i*p], &inc, &zero, vU, &inc);
      F77_NAME(daxpy)(&n, &one, y, &inc, vU, &inc);

      for(j = 0; j < n; j++){
	vU[j] *= 1.0/tauSq;
      }

      //L_A' b, the result is left in place 
      F77_NAME(dtrmv)(lower, yTran, nUnit, &n, L_A, &n, vU, &inc);
      
      //L_A (L_A'b) the result is left in place
      F77_NAME(dtrmv)(lower, nTran, nUnit, &n, L_A, &n, vU, &inc);
      
      //make z then do L_A z and save the result in the sample matrix column
      for(j = 0; j < n; j++){
    	REAL(wSamples_r)[i*n+j] = rnorm(0.0, 1.0);//z
      }

      //result left in place
      F77_NAME(dtrmv)(lower, nTran, nUnit, &n, L_A, &n, &REAL(wSamples_r)[i*n], &inc);
      
      //add L_A (L_A'b) to L_A z
      F77_NAME(daxpy)(&n, &one, vU, &inc, &REAL(wSamples_r)[i*n], &inc);

    }

     PutRNGstate();
 	
    //package beta and w samples into a list and ship it back to the calling program
    SEXP result_r, resultName_r;
    PROTECT(result_r = allocVector(VECSXP, 2)); nProtect++;
    PROTECT(resultName_r = allocVector(VECSXP, 2)); nProtect++;
    
    //samples
    SET_VECTOR_ELT(result_r, 0, betaSamples_r);
    SET_VECTOR_ELT(resultName_r, 0, mkChar("p.beta.samples")); 

    SET_VECTOR_ELT(result_r, 1, wSamples_r);
    SET_VECTOR_ELT(resultName_r, 1, mkChar("p.w.samples"));
    
    namesgets(result_r, resultName_r);
        
    //unprotect
    UNPROTECT(nProtect);
    
    return(result_r);
  }
}
